from dataclasses import dataclass
from .periods import Period
from datetime import datetime as DateTime
from typing import List, Optional, Any, Callable


@dataclass
class PolicyRule:
    applies_for: Period
    applies_period_count: int
    retain_every: Period


@dataclass
class _Result:
    index: int
    retain: bool
    time: DateTime


class RetentionPolicy:
    def __init__(self, layers: Optional[List[PolicyRule]] = None):
        self._layers = layers or []
        self._update()

    def _update(self):
        self._layers.sort(key=lambda l: l.applies_for.max_duration() * l.applies_period_count)

    def add_rule(self, applies_for: Period, applies_period_count: int, retain_every: Period):
        self._layers.append(PolicyRule(applies_for, applies_period_count, retain_every))
        self._update()

    def check_retention(self, items: List[Any],
                        key: Optional[Callable[[Any], DateTime]] = None,
                        now: Optional[DateTime] = None) -> List[bool]:
        """ Checks if the given time_stamps are to be retained according to the policy. """
        now = now or DateTime.now()

        # Construct the results
        key = key or (lambda x: x)
        results = [_Result(index, False, key(item) if key else now) for index, item in enumerate(items)]

        # We will iterate through each layer and mutate the results as we go. Layers will be able to mark a result as
        # to be retained, but will not un-mark something which has already been marked.
        for layer in self._layers:
            # Get the integer of the current layer's applies-to period
            now_period = layer.applies_for.to_period(now)

            # Find the items which are in the same applies-to period as the current time
            period_items = [item for item in results if
                            layer.applies_for.to_period(item.time) >= now_period - layer.applies_period_count + 1]

            # Now we will group the items by the retain-every period
            groups = {}
            for item in period_items:
                retain_period = layer.retain_every.to_period(item.time)
                if retain_period not in groups:
                    groups[retain_period] = []
                groups[retain_period].append(item)

            # Now we will iterate through the groups and mark the first item in each group as to be retained
            for group in groups.values():
                min(group, key=lambda x: x.time).retain = True

        # Finally, we will return the results
        return [x.retain for x in results]



"""
    Constructs for retention periods

    Retention periods are windows of time which are used to apply retention rules.  They are not simply unfixed
    durations of time, but instead are regions on a timeline.  For example, a "week" is not the same as "7 days", as
    the week is a period of time which starts on a specific day of the week and ends on the same day of the week.

    To handle this, retention periods must convert a date into an integer, where the next period can be found by
    incrementing the integer by one and the previous period can be found by decrementing it by one.  The easiest
    example of this is a "year", which is simply the year associated with the date.  There is a difference between
    specifying that one wants to retain data from each year, vs specifying that one wants to retain data from increments
    of time backwards from the present we usually associate with a year.

"""

from abc import ABC, abstractmethod
from datetime import datetime as DateTime, timedelta as TimeDelta

_reference_date = DateTime(1970, 1, 1)


def _year(time_stamp: DateTime) -> int:
    return time_stamp.year - _reference_date.year


def _day(time_stamp: DateTime) -> int:
    return int((time_stamp - _reference_date).days)


class Period(ABC):
    def to_period(self, time_stamp: DateTime) -> int:
        """Converts a time_stamp into an integer representing the period of time it is in"""
        raise NotImplementedError()

    def period_start(self, period: int) -> DateTime:
        """Returns the start time of the period"""
        raise NotImplementedError()

    def max_duration(self) -> TimeDelta:
        """Returns the maximum duration of the period"""
        raise NotImplementedError()


class SubdividedPeriod(Period):
    def __init__(self, sub_period: Period, subdivisions: int):
        self._sub_period = sub_period
        self._sub_div = subdivisions

    def to_period(self, time_stamp: DateTime) -> int:
        internal_period = self._sub_period.to_period(time_stamp)
        start_time = self._sub_period.period_start(internal_period)
        next_time = self._sub_period.period_start(internal_period + 1)
        fraction = (time_stamp - start_time) / (next_time - start_time)
        return internal_period * self._sub_div + int(fraction * self._sub_div)

    def period_start(self, period: int) -> DateTime:
        # Get the underlying period and its start
        p = period // self._sub_div
        internal_start = self._sub_period.period_start(p)
        internal_next = self._sub_period.period_start(p + 1)
        return internal_start + (internal_next - internal_start) * (period % self._sub_div) / self._sub_div

    def max_duration(self) -> TimeDelta:
        return self._sub_period.max_duration() / self._sub_div


class Year(Period):
    def to_period(self, time_stamp: DateTime) -> int:
        return _year(time_stamp)

    def max_duration(self) -> TimeDelta:
        return TimeDelta(days=366)

    def period_start(self, period: int) -> DateTime:
        return DateTime(_reference_date.year + period, 1, 1)


class Month(Period):
    def to_period(self, time_stamp: DateTime) -> int:
        return _year(time_stamp) * 12 + time_stamp.month

    def max_duration(self) -> TimeDelta:
        return TimeDelta(days=31)

    def period_start(self, period: int) -> DateTime:
        year = _reference_date.year + (period // 12)
        month = period % 12
        if month == 0:
            month = 12
            year -= 1
        return DateTime(year, month, 1)


class Week(Period):
    def to_period(self, time_stamp: DateTime) -> int:
        # This is easier than dealing with the years that have 53 weeks
        days = _day(time_stamp) - _reference_date.isocalendar()[2]
        return days // 7

    def max_duration(self) -> TimeDelta:
        return TimeDelta(days=7)

    def period_start(self, period: int) -> DateTime:
        return _reference_date + TimeDelta(days=(period * 7) + _reference_date.isocalendar()[2])


class Day(Period):
    def to_period(self, time_stamp: DateTime) -> int:
        return _day(time_stamp)

    def max_duration(self) -> TimeDelta:
        return TimeDelta(hours=24)

    def period_start(self, period: int) -> DateTime:
        return _reference_date + TimeDelta(days=period)


class Hour(Period):
    def to_period(self, time_stamp: DateTime) -> int:
        return _day(time_stamp) * 24 + time_stamp.hour

    def max_duration(self) -> TimeDelta:
        return TimeDelta(minutes=60)

    def period_start(self, period: int) -> DateTime:
        return _reference_date + TimeDelta(hours=period)


class Minute(Period):
    def to_period(self, time_stamp: DateTime) -> int:
        return (_day(time_stamp) * 24 + time_stamp.hour) * 60 + time_stamp.minute

    def max_duration(self) -> TimeDelta:
        return TimeDelta(seconds=60)

    def period_start(self, period: int) -> DateTime:
        return _reference_date + TimeDelta(minutes=period)



"""
    These tests confirm that the period classes perform as expected by brute force testing them over wide ranges of
    dates.  Rather than look directly for expected values, we are checking for the following behaviors:

    * When advancing a clock by time stamps that are significantly smaller than the rough duration of the
        period (keeping in mind that periods like "week" and "month" are not fixed durations), the period integer
        will only ever increase by one OR stay the same.

    * There is a duration which represents the longest possible instance of the period, and that after the clock has
        advanced by that duration, the period integer MUST have increased by one.  For example, after 366 days, the
        year integer MUST have increased by one.

    Because of this strategy, these tests are relatively slow to run and may take several seconds to complete.
    However, they should detect edge cases which might otherwise slip through.

"""

from datetime import datetime as DateTime, timedelta as TimeDelta
from retention_rules.periods import Year, Month, Week, Day, Hour, Minute, Period, SubdividedPeriod


def test_month_advancement():
    """ Tests that the month period advances correctly. """
    tester = AdvancementTester(Month(), TimeDelta(hours=4), TimeDelta(days=31))
    for t, m in tester.test_between(DateTime(1980, 1, 1), DateTime(2040, 1, 1)):
        assert t, m


def test_year_advancement():
    """ Tests that the year period advances correctly. """
    tester = AdvancementTester(Year(), TimeDelta(hours=4), TimeDelta(days=366))  # Leap years have 366 days
    for t, m in tester.test_between(DateTime(1980, 1, 1), DateTime(2040, 1, 1)):
        assert t, m


def test_week_advancement():
    """ Tests that the week period advances correctly. """
    tester = AdvancementTester(Week(), TimeDelta(hours=4), TimeDelta(days=7))
    for t, m in tester.test_between(DateTime(1980, 1, 1), DateTime(2040, 1, 1)):
        assert t, m


def test_week_rollover():
    """ Tests that the week period rolls over correctly. """
    week = Week()
    clock = DateTime(2020, 1, 4)
    last_clock = clock
    last_period = week.to_period(clock)
    while clock < DateTime(2035, 1, 1):
        period = week.to_period(clock)
        if clock.isocalendar()[2] == 1:
            assert period - last_period == 1, f"Week should have advanced by one from {last_clock} -> {clock}"

        last_clock = clock
        last_period = period
        clock += TimeDelta(days=1)


def test_day_advancement():
    """ Tests that the day period advances correctly. """
    tester = AdvancementTester(Day(), TimeDelta(hours=1), TimeDelta(hours=24))
    for t, m in tester.test_between(DateTime(2020, 1, 1), DateTime(2040, 1, 1)):
        assert t, m


def test_hour_advancement():
    """ Tests that the hour period advances correctly. """
    tester = AdvancementTester(Hour(), TimeDelta(minutes=6), TimeDelta(hours=1))
    for t, m in tester.test_between(DateTime(2020, 1, 1), DateTime(2026, 1, 1)):
        assert t, m


def test_minute_advancement():
    """ Tests that the minute period advances correctly. """
    tester = AdvancementTester(Minute(), TimeDelta(seconds=30), TimeDelta(minutes=1))
    for t, m in tester.test_between(DateTime(2024, 1, 1), DateTime(2024, 12, 31, 23, 59, 59)):
        assert t, m


def test_subdivided_hours():
    period = SubdividedPeriod(Hour(), 4)
    tester = AdvancementTester(period, TimeDelta(minutes=3), TimeDelta(minutes=15))
    for t, m in tester.test_between(DateTime(2024, 1, 1), DateTime(2025, 1, 1)):
        assert t, m


class AdvancementTester:
    def __init__(self, period: Period, advance_by: TimeDelta, verify_increment: TimeDelta):
        self.period = period
        self.advance_by = advance_by
        self.verify_increment = verify_increment

    def test_between(self, start: DateTime, end: DateTime):
        """ Tests that the period advances correctly between two dates """
        clock = start
        last_clock = clock
        last_period = self.period.to_period(clock)
        last_change = clock

        while clock < end:
            clock += self.advance_by
            period = self.period.to_period(clock)

            yield 0 <= period - last_period <= 1, f"Periods should only advance by one at a time, but {last_clock} " \
                                                  f"-> {clock} advanced by {period - last_period}"

            if period - last_period == 1:
                last_change = clock

            if clock - last_change > self.verify_increment:
                yield False, f"Period should have advanced between {last_change} -> {clock}, but did not ({period})"

            last_clock = clock
            last_period = period



"""
    Tests to check the policy building
"""
from typing import Tuple

from retention_rules.builder import build_policy, parse_period_text
from retention_rules.periods import Period, Year, Month, Week, Day, Hour, Minute, SubdividedPeriod


def test_parse_simple():
    assert parse_period_text("Y") == _Check(Year(), 1)
    assert parse_period_text("M") == _Check(Month(), 1)
    assert parse_period_text("W") == _Check(Week(), 1)
    assert parse_period_text("D") == _Check(Day(), 1)
    assert parse_period_text("H") == _Check(Hour(), 1)
    assert parse_period_text("MIN") == _Check(Minute(), 1)


def test_parse_with_count():
    assert parse_period_text("2Y") == _Check(Year(), 2)
    assert parse_period_text("2M") == _Check(Month(), 2)
    assert parse_period_text("2W") == _Check(Week(), 2)
    assert parse_period_text("2D") == _Check(Day(), 2)
    assert parse_period_text("2H") == _Check(Hour(), 2)
    assert parse_period_text("2MIN") == _Check(Minute(), 2)


def test_parse_with_sub_div():
    assert parse_period_text("Y/2") == _Check(SubdividedPeriod(Year(), 2), 1)
    assert parse_period_text("M/2") == _Check(SubdividedPeriod(Month(), 2), 1)
    assert parse_period_text("W/2") == _Check(SubdividedPeriod(Week(), 2), 1)
    assert parse_period_text("D/2") == _Check(SubdividedPeriod(Day(), 2), 1)
    assert parse_period_text("H/2") == _Check(SubdividedPeriod(Hour(), 2), 1)
    assert parse_period_text("MIN/2") == _Check(SubdividedPeriod(Minute(), 2), 1)


def test_parse_with_count_and_sub_div():
    assert parse_period_text("3Y/2") == _Check(SubdividedPeriod(Year(), 2), 3)
    assert parse_period_text("3M/2") == _Check(SubdividedPeriod(Month(), 2), 3)
    assert parse_period_text("3W/2") == _Check(SubdividedPeriod(Week(), 2), 3)
    assert parse_period_text("3D/2") == _Check(SubdividedPeriod(Day(), 2), 3)
    assert parse_period_text("3H/2") == _Check(SubdividedPeriod(Hour(), 2), 3)
    assert parse_period_text("3MIN/2") == _Check(SubdividedPeriod(Minute(), 2), 3)


class _Check:
    def __init__(self, period, count: int):
        self.period = period
        self.count = count

    def __eq__(self, other: Tuple[Period, int]):
        if isinstance(other[0], SubdividedPeriod):
            if not isinstance(self.period, SubdividedPeriod):
                return False
            return (isinstance(other[0]._sub_period, type(self.period._sub_period))
                    and other[0]._sub_div == self.period._sub_div
                    and other[1] == self.count)
        return isinstance(other[0], type(self.period)) and other[1] == self.count



"""
    This module converts python dictionaries into a policy object.

    A dictionary will look something like this:

    {
        "rules": [
            {"applies_for": "3D", "retain_every": "H/4"},
            {"applies_for": "2W", "retain_every": "H"},
            {"applies_for": "M", "retain_every": "D/2"},
            {"applies_for": "6M", "retain_every": "W/2"},
            {"applies_for": "Y", "retain_every": "W"},
            {"applies_for": "10Y", "retain_every": "M"},
        ]
    }

"""

import re
from typing import Dict, Tuple

from .policy import RetentionPolicy
from .periods import *


def build_policy(policy_dict: Dict) -> RetentionPolicy:
    policy = RetentionPolicy()
    rule_items = policy_dict.get("rules", [])
    if not rule_items:
        raise ValueError("Policy must contain at least one rule in the 'rules' key")

    for rule in rule_items:
        applies, count = parse_period_text(rule["applies_for"])
        retain_every, rcount = parse_period_text(rule["retain_every"])
        if rcount != 1:
            raise ValueError(f"Retain-every does not currently support a count")

        policy.add_rule(applies, count, retain_every)
    return policy


_key_pattern = re.compile(r"^(\d+)?(Y|M|W|D|H|MIN|S)(/\d+)?$")

_by_key = {
    "Y": Year,
    "M": Month,
    "W": Week,
    "D": Day,
    "H": Hour,
    "MIN": Minute,
}


def parse_period_text(text: str) -> Tuple[Period, int]:
    match = _key_pattern.match(text)
    if not match:
        raise ValueError(f"Invalid period: {text}")

    count = int(match.group(1) or 1)
    unit = match.group(2)
    sub_div = match.group(3)

    period_cls = _by_key.get(unit, None)
    if not period_cls:
        raise ValueError(f"Could not find period for unit: {unit}")

    period = period_cls()
    if sub_div:
        period = SubdividedPeriod(period, int(sub_div[1:]))
    return period, count



from datetime import datetime as DateTime, timedelta as TimeDelta
from retention_rules.periods import Year, Month, Week, Day, Hour, Minute, Period, SubdividedPeriod


def test_year_start_times():
    """ Tests that the year period starts at the correct times. """
    year = Year()
    for i in range(100):
        start = year.period_start(i)
        assert i == year.to_period(start), f"Year {i} ({start}) failed round trip"
        assert i - 1 == year.to_period(start - TimeDelta(seconds=1)), f"Year {i} ({start}) failed to decrement"


def test_month_start_times():
    """ Tests that the month period starts at the correct times. """
    month = Month()
    offset = month.to_period(DateTime(1980, 1, 1))

    for j in range(1200):
        i = j + offset
        start = month.period_start(i)
        assert i == month.to_period(start), f"Month {i} ({start}) failed round trip"
        assert i - 1 == month.to_period(start - TimeDelta(seconds=1)), f"Month {i} ({start}) failed to decrement"


def test_week_start_times():
    """ Tests that the week period starts at the correct times. """
    week = Week()
    offset = week.to_period(DateTime(2000, 1, 1))

    for j in range(52 * 50):
        i = j + offset
        start = week.period_start(i)
        assert i == week.to_period(start), f"Week {i} ({start}) failed round trip"
        assert i - 1 == week.to_period(start - TimeDelta(seconds=1)), f"Week {i} ({start}) failed to decrement"


def test_day_start_times():
    """ Tests that the day period starts at the correct times. """
    day = Day()
    offset = day.to_period(DateTime(2020, 1, 1))

    for j in range(365 * 50):
        i = j + offset
        start = day.period_start(i)
        assert i == day.to_period(start), f"Day {i} ({start}) failed round trip"
        assert i - 1 == day.to_period(start - TimeDelta(seconds=1)), f"Day {i} ({start}) failed to decrement"


def test_hour_start_times():
    """ Tests that the hour period starts at the correct times. """
    hour = Hour()
    offset = hour.to_period(DateTime(2015, 1, 1))

    for j in range(24 * 365 * 10):
        i = j + offset
        start = hour.period_start(i)
        assert i == hour.to_period(start), f"Hour {i} ({start}) failed round trip"
        assert i - 1 == hour.to_period(start - TimeDelta(seconds=1)), f"Hour {i} ({start}) failed to decrement"


def test_minute_start_times():
    """ Tests that the minute period starts at the correct times. """
    minute = Minute()
    offset = minute.to_period(DateTime(2024, 1, 1))

    for j in range(60 * 24 * 200):
        i = j + offset
        start = minute.period_start(i)
        assert i == minute.to_period(start), f"Minute {i} ({start}) failed round trip"
        assert i - 1 == minute.to_period(start - TimeDelta(seconds=1)), f"Minute {i} ({start}) failed to decrement"


def test_subdivided_hour_start_times():
    period = SubdividedPeriod(Hour(), 4)
    offset = period.to_period(DateTime(2015, 1, 1))

    for j in range(24 * 4 * 365 * 2):
        i = j + offset
        start = period.period_start(i)
        assert start.minute in [0, 15, 30, 45], f"Minute {start.minute} is not a quarter hour"
        assert i == period.to_period(start), f"Quarter hour {i} ({start}) failed round trip"
        assert i - 1 == period.to_period(start - TimeDelta(seconds=1)), f"Quarter hour {i} ({start}) failed to decrement"
